@@ -1,9 +1,19 @@
 "Module providing login utilities for twspace_dl"
 import re
+from datetime import datetime, timedelta
 
 import requests
 
 
+def is_expired(filename: str) -> bool:
+    """return whether a cookie is expired"""
+    timestamp = re.findall(
+        r"\d{10}(?=\sauth_token\s)", open(filename, "r", encoding="utf-8").read()
+    )[0]
+    date = datetime.fromtimestamp(float(timestamp))
+    return date < datetime.now()
+
+
 def load_from_file(filename: str) -> str:
     """return auth_token from netscape cookie file"""
     try:
@@ -22,12 +32,18 @@ def load_from_file(filename: str) -> str:
 
 def write_to_file(auth_token: str, filename: str) -> None:
     """Write cookie to file in a format recognizable by the module
-    (`auth_token        {auth_token}`)"""
+    (`{timestamp}\tauth_token\t{auth_token}`)"""
+    # Taken from twitter's set-cookie parameter(13 months)
+    delta = timedelta(seconds=34214400)
+    expiry_date = datetime.now() + delta
+    timestamp = int(expiry_date.timestamp())
     with open(filename, "w", encoding="utf-8") as cookie_file:
-        cookie_file.write("auth_token" + 8 * " " + auth_token)
+        cookie_file.write(f"{timestamp}\tauth_token\t{auth_token}")
 
 
 class Login:
+    """Helper class to login to twitter"""
+
     def __init__(self, username, password, guest_token):
         self.username = username
         self.password = password
@@ -37,6 +53,7 @@ class Login:
         self.flow_token: str
 
     def login(self) -> str:
+        """Login to twitter"""
         request_flow = self.session.post(
             self.task_url,
             params={"flow_name": "login"},
@@ -70,29 +87,54 @@ class Login:
         except KeyError as err:
             raise RuntimeError("Error identifying user:", request_flow.json()) from err
 
-        # account duplication check
+        # alternate identifier
         request_flow = self.session.post(
-            self.task_url, headers=self._headers, json=self._account_dup_check_data
-        )
-        try:
-            self.flow_token = request_flow.json()["flow_token"]
-        except KeyError as err:
-            raise RuntimeError(
-                "Error while checking account duplication:", request_flow.json()
-            ) from err
+            self.task_url, headers=self._headers, json=self._login_alternate_identifier
+        ).json()
+        if "flow_token" in request_flow.keys():
+            self.flow_token = request_flow["flow_token"]
+            # Sometimes it doesn't check for alternate id
+            # raise RuntimeError(
+            #     "Error while checking for alternate id", request_flow.json()
+            # ) from err
 
         # enter password
         request_flow = self.session.post(
             self.task_url, headers=self._headers, json=self._enter_password_data
         )
-        try:
-            auth_token = str(request_flow.cookies["auth_token"])
-        except KeyError as err:
+        if "auth_token" in request_flow.cookies.keys():
+            return str(request_flow.cookies["auth_token"])
+        if (
+            "LoginTwoFactorAuthChallenge" in request_flow.text
+            or "AccountDuplicationCheck" in request_flow.text
+        ):
+            self.flow_token = request_flow.json()["flow_token"]
+        else:
             raise RuntimeError(
                 "Error while while entering password:", request_flow.json()
-            ) from err
+            )
 
-        return auth_token
+        # account duplication check
+        request_flow = self.session.post(
+            self.task_url, headers=self._headers, json=self._account_dup_check_data
+        ).json()
+        if "auth_token" in request_flow.cookies.keys():
+            return str(request_flow.cookies["auth_token"])
+        if "flow_token" in request_flow.keys():
+            self.flow_token = request_flow["flow_token"]
+            # Sometimes it doesn't check account duplication
+            # raise RuntimeError(
+            #     "Error while checking account duplication:", request_flow.json()
+            # ) from err
+
+        # 2FA
+        request_flow = self.session.post(
+            self.task_url, headers=self._headers, json=self._enter_2fa
+        )
+        if "auth_token" in request_flow.cookies.keys():
+            print("Success!")
+            return str(request_flow.cookies["auth_token"])
+        raise RuntimeError("Login failed after 2FA. Error message: ", request_flow.text)
 
     @property
     def _headers(self):
@@ -105,6 +147,24 @@ class Login:
             "x-guest-token": self.guest_token,
         }
 
+    @property
+    def _initial_params(self) -> dict:
+        return {
+            "input_flow_data": {
+                "flow_context": {
+                    "debug_overrides": {},
+                    "start_location": {"location": "splash_screen"},
+                }
+            },
+            "subtask_versions": {
+                "contacts_live_sync_permission_prompt": 0,
+                "email_verification": 1,
+                "topics_selector": 1,
+                "wait_spinner": 1,
+                "cta": 4,
+            },
+        }
+
     @property
     def _js_instrumentation_data(self) -> dict:
         return {
@@ -159,6 +219,18 @@ class Login:
             ],
         }
 
+    @property
+    def _login_alternate_identifier(self) -> dict:
+        return {
+            "flow_token": self.flow_token,
+            "subtask_inputs": [
+                {
+                    "subtask_id": "LoginEnterAlternateIdentifierSubtask",
+                    "enter_text": {"text": self.username, "link": "next_link"},
+                }
+            ],
+        }
+
     @property
     def _account_dup_check_data(self) -> dict:
         # assert self.flow_token[-1] == "2"
@@ -188,19 +260,17 @@ class Login:
         }
 
     @property
-    def _initial_params(self) -> dict:
+    def _enter_2fa(self) -> dict:
+        # assert self.flow_token[-1] == "6"
         return {
-            "input_flow_data": {
-                "flow_context": {
-                    "debug_overrides": {},
-                    "start_location": {"location": "splash_screen"},
+            "flow_token": self.flow_token,
+            "subtask_inputs": [
+                {
+                    "subtask_id": "LoginTwoFactorAuthChallenge",
+                    "enter_text": {
+                        "text": input("Enter 2FA code:\t"),
+                        "link": "next_link",
+                    },
                 }
-            },
-            "subtask_versions": {
-                "contacts_live_sync_permission_prompt": 0,
-                "email_verification": 1,
-                "topics_selector": 1,
-                "wait_spinner": 1,
-                "cta": 4,
-            },
+            ],
         }
